/**
    bespoke synth, a software modular synthesizer
    Copyright (C) 2021 Ryan Challinor (contact: awwbees@gmail.com)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
**/
//
//  IAudioSource.cpp
//  Bespoke
//
//  Created by Ryan Challinor on 12/16/15.
//
//

#include "IAudioSource.h"
#include "IAudioReceiver.h"
#include "PatchCableSource.h"

IAudioReceiver* IAudioSource::GetTarget(int index)
{
   assert(index < GetNumTargets());
   return GetPatchCableSource(index)->GetAudioReceiver();
}

void IAudioSource::SyncOutputBuffer(int numChannels)
{
   for (int i = 0; i < GetNumTargets(); ++i)
   {
      if (GetTarget(i))
      {
         ChannelBuffer* out = GetTarget(i)->GetBuffer();
         out->SetNumActiveChannels(MAX(numChannels, out->NumActiveChannels()));
      }
   }
   GetVizBuffer()->SetNumChannels(numChannels);
}
