#include "VersionInfo.h"

// This file will get replaced with the CMAKE variables as below at build time
namespace Bespoke
{

    const char* GIT_BRANCH = "makepkg";
    const char* GIT_HASH = "1d0f0429";
    const char* BUILD_ARCH = "x86_64";
    const char* BUILD_DATE = "2026-03-26";
    const char* BUILD_TIME = "18:30:53";

}
