#include "VersionInfo.h"

// This file will get replaced with the CMAKE variables as below at build time
namespace Bespoke
{
    const char* APP_NAME = "BespokeSynth";
    const char* VERSION = "1.3.0";
    const char* PYTHON_VERSION ="3.14.3";
    const char* CMAKE_INSTALL_PREFIX = "/usr";
}
